import logging
import re

from .connection import DSMConnection
from .process import Process

LOG = logging.getLogger(__name__)


class Task():
    """
    Reimplementation of the shell scripts on https://git.uis.cam.ac.uk/i/uis/infra/dellsc.git
    """
    def __init__(self, host, port, user, password, verify):
        self.process = Process()
        self.dsm_connection = DSMConnection(host, port, user, password, verify)

    def __enter__(self):
        self.dsm_connection.open_connection()
        return self

    def __exit__(self, tipe, value, traceback):
        self.dsm_connection.close_connection()
        self.dsm_connection = None
        self.process = None

    def get_volume_info(self, wwn=None, mountpoint=None):
        """
        Returns details of volume with WWN `wwn` or mounted on `mountpoint`.

        :param wwn: volume's `wwn`.
        :param mountpoint: `mountpoint` where the volume is mounted.
        :returns: tuple (wwn, volid, volname, volFolderName, volFolderId, scName), `None`, otherwise.
        """
        if wwn is None and mountpoint is None:
            return None

        if mountpoint is not None:
            mpath = self.process.get_mpath_from_mountpoint(mountpoint)
            if mpath is None:
                return None

            wwn = self.process.get_wwn_from_mpath(mpath)
            if wwn is None:
                return None

        # get list of volumes
        self.dsm_connection.get_volumes()

        if self.dsm_connection.volumes.get(wwn) is not None:
            volid = self.dsm_connection.volumes[wwn].get('instanceId')
            volname = self.dsm_connection.volumes[wwn].get('name')
            volFolderName = self.dsm_connection.volumes[wwn].get('volumeFolderName')
            volFolderId = self.dsm_connection.volumes[wwn].get('volumeFolderId')
            volScName = self.dsm_connection.volumes[wwn].get('scName')
        else:
            LOG.error('Cannot get volume with WWN %s.', wwn)
            return None

        return (wwn, volid, volname, volFolderName, volFolderId, volScName)

    def get_vol_wwn_from_id(self, volid):
        """
        Returns volume wwn from volume ID.

        :param volid: volume ID.
        :return: volume WWN, `None` otherwise.
        """
        self.dsm_connection.get_volumes()
        for wwn, vol_info in self.dsm_connection.volumes.items():
            if self.dsm_connection.volumes[wwn]['instanceId'] == str(volid):
                return wwn

        LOG.error('Cannot get the volume WWN from the volume ID %s.', volid)

        return None

    def generate_clone_name(self, volname):
        """
        Generates a clone name from a volume name.

        :param volname: the volume name from which to generate a clone name.
        :returns: clone name, `None`, otherwise.
        """
        # remove clone name suffix from volume name
        orig_volname = re.sub(r'_\d+$', '', volname)

        # get list of volumes starting with 'orig_volname'
        self.dsm_connection.get_volumes()
        lst = []
        for vol in self.dsm_connection.volumes.values():
            if vol['name'].startswith(orig_volname) and not vol['inRecycleBin']:
                lst.append(vol['name'])

        if len(lst) == 0:
            LOG.error('Clone name cannot be generated as no volumes is found.')
            return None

        # get indexes in the volume list
        volname_with_index = re.compile(r'.*(_\d+)$')
        volindex_lst = [int(volname_.replace(f'{orig_volname}_', ''))
                        for volname_ in lst if volname_with_index.match(volname_)]
        volindex_lst.sort()
        if len(volindex_lst) == 0:
            clone_index = 1
        else:
            clone_index = volindex_lst[-1] + 1

        return f'{orig_volname}_{clone_index}'

    def get_local_server_id(self):
        """Returns local server ID, or `None` on error."""
        hostname = self.process.get_hostname()
        server_name = hostname.split('.')[0]

        server_id = self.dsm_connection.get_server_id(server_name)
        if server_id is None:
            LOG.error('Cannot get server ID of local server %s.', server_name)

        return server_id

    def replace_volume(self, src_mountpoint, dst_mountpoint):
        """
        Clones the volume mounted on `src_mountpoint` and mount it on `dst_mountpoint`,
        the volume that was mounted on `dst_mountpoint` will be deleted.

        :param src_mountpoint: mountpoint of the volume to be cloned.
        :param dst_mountpoint: mountpoint of the clone volume.
        :return: `True` for success, `False` on error.
        """
        # check details of vols on source and destination mountpoints
        # volume details: (wwn, volid, volname, volFolderName, volFolderId, scName)
        src_vol_info = self.get_volume_info(mountpoint=src_mountpoint)
        if src_vol_info is None:
            return False

        dst_vol_info = self.get_volume_info(mountpoint=dst_mountpoint)
        if dst_vol_info is None:
            return False

        # unmount the volume to be cloned, call it "the source volume"
        if not self.process.umount(src_mountpoint):
            return False

        # create a volume snapshot from the source volume
        snapshot_id = self.dsm_connection.create_snapshot(src_vol_info[1])
        if snapshot_id is None:
            return False

        # check if the source volume is a camsis volume
        replay_profile_id = None
        self.dsm_connection.get_snapshot_profiles()
        if self.dsm_connection.rply_profiles is None:
            return False

        if 'camsis' in src_vol_info[3].lower():
            # we're sure 'camsis' replay profile exists
            replay_profile_id = self.dsm_connection.rply_profiles['camsis']['instanceId']

        # create a clone volume from the source volume
        clone_name = self.generate_clone_name(src_vol_info[2])
        view_vol_id = self.dsm_connection.create_view_volume(
            snapshot_id, replay_profile_id, clone_name, src_vol_info[4])

        # get local server id and map view volume to server
        server_id = self.get_local_server_id()
        if server_id is None:
            return False

        if not self.dsm_connection.map_volume_to_server(view_vol_id, server_id):
            return False

        # mount the source volume
        if not self.process.mount(src_mountpoint):
            return False

        # prepare to mount the view volume in place of the destination volume
        view_vol_wwn = self.get_vol_wwn_from_id(view_vol_id)
        if view_vol_wwn is None:
            return False

        # rescan scsi bus to read the new volume mapped to the server
        if not self.process.rescan_scsibus():
            return False

        # update multipath.conf
        if not self.process.add_multipath_alias(view_vol_wwn):
            return False

        dst_vol_fs_device = self.process.get_fsdevice_from_mountpoint(dst_mountpoint)
        view_vol_alias = self.process.get_mpath_from_wwn(view_vol_wwn)
        view_vol_fs_device = f'/dev/mapper/{view_vol_alias}p1'

        # unmount the destination volume and mount view volume
        if not self.process.umount(dst_mountpoint):
            return False

        if not self.process.replace_entry_in_fstab(dst_vol_fs_device, view_vol_fs_device):
            return False

        if not self.process.daemon_reload():
            return False

        if not self.process.mount(dst_mountpoint):
            return False

        # delete destination volume
        result_delete_vol = self.delete_volume(dst_vol_info[0])

        return True if result_delete_vol else False

    def unmap_dellsc_volume(self, wwn):
        """
        Unmaps a Dell SC volume with `WWN`. Unmounting the volume should be performed manually.

        :param wwn: `wwn` of the volume to unmap.
        :returns: `True` for success, `False` on error.
        """
        # check if volume exists on Dell SCs
        vol_info = self.get_volume_info(wwn)
        if vol_info is None:
            return False

        # check if volume is visible to local server
        alias = self.process.get_mpath_from_wwn(wwn)
        if alias is None:
            return False

        # make sure volume is mapped to local server
        vol_mappings = self.dsm_connection.get_vol_mappings(vol_info[1])
        if vol_mappings is None:
            return False

        servers = [x['server']['instanceName'] for x in vol_mappings]
        if self.process.get_hostname().split('.')[0] not in servers:
            LOG.error('Volume with ID %s is not mapped to local server.', vol_info[1])
            return False

        # make sure volume is not mounted
        if self.process.is_wwn_mounted(wwn) is not None:
            LOG.error('Volume with WWN %s is mounted on local server. Unmount it first.', wwn)
            return False

        # remove volume from fstab config file
        device_name = f'/dev/mapper/{alias}p1'
        if self.process.is_device_in_fstab(device_name):
            if not self.process.remove_entry_from_fstab(device_name):
                return False
            if not self.process.daemon_reload():
                return False
        else:
            LOG.warning('%s is not in /etc/fstab.', device_name)

        # remove (wwid, alias) entry from multipath.conf
        if not self.process.remove_multipath_entry(wwn, alias):
            return False

        # delete raw devices of multipath device alias
        raw_devices = self.process.get_multipath_raw_devices(alias)
        if raw_devices is None:
            LOG.warning('No raw devices for multipath device %s.', alias)
        else:
            result_delete_rawdevs = []
            for rawdev in raw_devices:
                result = self.process.delete_raw_devices(rawdev)
                result_delete_rawdevs.append(result)
            if False in result_delete_rawdevs:
                return False

        # reload multipathd
        if not self.process.reload_multipathd():
            return False

        # unmap the volume
        if not self.dsm_connection.unmap_volume(vol_info[1]):
            return False

        return False

    def delete_volume(self, wwn, recycle=True):
        """
        Deletes a volume with `wwn`.

        :param wwn: `wwn` of the volume to delete.
        :returns: `True` for success, `False` on error.
        """
        # unmap the volume from the local server
        if not self.unmap_dellsc_volume(wwn):
            return False

        # get volid
        vol_info = self.get_volume_info(wwn)
        if vol_info is None:
            return False

        # delete the volume
        if not self.dsm_connection.delete_volume(vol_info[1], recycle=recycle):
            return False

        return True

    def map_dellsc_volume(self, wwn, mountpoint):
        """
        Maps Dell SC volume with `WWN` to local server and mounts it on `mountpoint`.

        :param wwn: WWN of the volume to map to local server.
        :param mountpoint: mountpoint of the volume.
        :returns: `True` for success, `False` on error.
        """
        # check volume exists and get its ID
        vol_info = self.get_volume_info(wwn)
        if vol_info is None:
            return False

        # check if the volume is already mapped to local server or other servers
        vol_mappings = self.dsm_connection.get_vol_mappings(vol_info[1])
        if vol_mappings is None:
            return False

        if len(vol_mappings) > 0:
            LOG.error('Volume %s is already mapped to: %s.',
                      wwn, [x['server']['instanceName'] for x in vol_mappings])
            return False

        # get ID of local server id and map view volume to it
        server_id = self.get_local_server_id()
        if server_id is None:
            return False

        if not self.dsm_connection.map_volume_to_server(vol_info[1], server_id):
            return False

        # rescan scsi bus
        if not self.process.rescan_scsibus():
            return False

        # add multipath alias of view volume wwid
        if not self.process.add_multipath_alias(wwn):
            return False

        # update fstab file
        new_alias = self.process.get_mpath_from_wwn(wwn)
        if new_alias is not None:
            new_fsdevice = f'/dev/mapper/{new_alias}p1'
            if not self.process.add_entry_in_fstab(new_fsdevice, mountpoint):
                LOG.error('%s cannot be added to /etc/fstab', new_fsdevice)
                return False
            if not self.process.daemon_reload():
                return False
        else:
            LOG.error('Alias for the view volume WWN %s does not exist.', mountpoint)
            return False

        # mount target_mountpoint if it's not mounted
        # found it gets mounted after reload multipath, is it an expected behaviour?
        if not self.process.mountpoint_exists(mountpoint):
            if not self.process.mount(mountpoint):
                return False

        return True

    def clone_volume(self, volume_wwn, replay_label, target_mountpoint):
        """
        Creates a snapshot of a volume, then creates a view volume from the snapshot,
        map the view volume to local server and mount it on `target_mountpoint`.

        :param volume_wwn: wwn of volume from which to create the snapshot and the view volume.
        :param replay_label: label of the snapshot to create the volume from. Also it will be used
        to name the new volume.
        :param target_mounpoint: where to mount the view volume.
        :return: `True` for success, `False` on failure.
        """
        if not self.process.is_dir_empty(target_mountpoint):
            return False

        # get volume info: (wwn, volid, volname, volFolderName, volFolderId, scName)
        vol_info = self.get_volume_info(wwn=volume_wwn)
        if vol_info is None:
            return False

        # create snapshot from volume
        snapshot_id = self.dsm_connection.create_snapshot(vol_info[1], description=replay_label)
        if snapshot_id is None:
            return False

        # create view volume
        view_volume_name = f"{vol_info[2]}_{replay_label}"
        view_volume_id = self.dsm_connection.create_view_volume(
            snapshot_id, None, view_volume_name, vol_info[4])

        # get view volume wwn
        view_volume_wwn = self.get_vol_wwn_from_id(view_volume_id)

        if view_volume_wwn is None:
            return False

        # map the volume
        if not self.map_dellsc_volume(self, view_volume_wwn, target_mountpoint):
            return False

        return True

    def create_snapshot(self, mountpoint, label, retention):
        """
        Creates a snapshot of the volume mounted on `mountpoint`.

        :param mountpoint: the mountpoint where the volume is mounted.
        :param label: snapshot label.
        :param retention: retention period of the snapshot to create, in days.
        :return: `True` for success, `False` on error.
        """
        # check if snapshot label is valid
        if not self.process.is_snapshot_label_valid(label):
            LOG.error('Snapshot label %s is not valid.', label)
            return False

        # check if the retention is between 1 and 99 days
        if not (int(retention) in range(1, 100)):
            LOG.error('Retention %s is not valid.', retention)
            return False

        # check if volume is on DSM
        vol_info = self.get_volume_info(mountpoint=mountpoint)
        if vol_info is None:
            LOG.error('No info on volume mounted on %s.', mountpoint)
            return False

        # get fsdevice. TODO: is this check really needed?
        fsdevice = self.process.get_fsdevice_from_mountpoint(mountpoint)
        if fsdevice is None:
            LOG.error('No fs device of volume mounted on %s.', mountpoint)
            return False

        # create the snapshot
        result = self.dsm_connection.create_snapshot(vol_info[1], label, int(retention)*1440)

        return False if result is None else True

    def delete_recycled_vols(self, recycled_vols_file):
        """
        Deletes the volumes listed in a file passed in argument.
        The volume must be in Recycle Bin to be deleted.

        :param recycled_vols_file: file containing the list of recycled volumes to be deleted.
        Each line in the file should contain the volume's WWN.
        The file should only contain lines in the following format:
        ('6000d31000e3940000000000000001c9', {'instanceId': '58260.165', \
            'inRecycleBin': True, 'volumeFolderPath': 'All Volumes/SB_VMWare_vols/', \
            'volumeFolderName': 'SB_VMWare_vols', 'volumeFolderId': '58260.11', \
            'status': 'Down', 'name': 'test_vol', 'scName': 'WCDC Storage Center'})
        :return: `True` if all the volumes that are in Recycle Bin are deleted, `False` otherwise.
        """
        if not self.dsm_connection.get_recycled_volumes():
            return False

        wwns = []
        with open(recycled_vols_file, 'r') as rvf:
            # extract the wwn and ignore the commented lines
            for linenum, line in enumerate(rvf):
                if not line.startswith('#'):
                    rs = re.search(r"\('[A-Za-z0-9]{32}'", line)
                    if rs:
                        wwns.append(rs.group(0).replace("'", '').replace("(", ''))
                    else:
                        LOG.warning("Cannot find a valid WWN in line number %s (%s...)", linenum, line[0:50])

        results = []
        for wwn in wwns:
            if self.dsm_connection.recycled_vols.get(wwn) is None:
                LOG.warning(f"Volume with WWN '{wwn}' will not be deleted as the WWN may not be valid"
                            " or the volume is not in Recycle Bin.")
            else:
                LOG.info('Volume %s will be DELETED from the Recyle Bin!', wwn)
                result_del_vol = self.dsm_connection.delete_volume(
                    self.dsm_connection.recycled_vols[wwn]['instanceId'], recycle=False)
                results.append(result_del_vol)

        return False if False in results else True



import json
import requests

API_VERSION = '2.0'


class HttpClient:
    """Wrapper class to make Storage Center REST API calls."""

    def __init__(self, host, port, user, password, verify):
        """
        HttpClient handles the REST requests.

        :param host: IP address of the Dell Data Collector.
        :param port: Port the Data Collector is listening on.
        :param user: User account to login with.
        :param password: Password.
        :param verify: Boolean indicating whether certificate verification
                       should be turned on or not.
        """
        self.base_url = 'https://%s:%s/api/rest/' % (host, port)
        self.session = requests.Session()
        self.session.auth = (user, password)

        self.header = {}
        self.header['Content-Type'] = 'application/json; charset=utf-8'
        self.header['Accept'] = 'application/json'
        self.header['x-dell-api-version'] = API_VERSION
        self.verify = verify

        if not verify:
            requests.packages.urllib3.disable_warnings()

    def __enter__(self):
        return self

    def __exit__(self):
        self.session.close()

    def _format_url(self, url):
        """Formats the REST URL to use for API calls."""
        return '%s%s' % (self.base_url, url if url[0] != '/' else url[1:])

    def get(self, url):
        """Perform a REST GET request."""
        return self.session.get(
            self._format_url(url),
            headers=self.header,
            verify=self.verify)

    def post(self, url, payload):
        """Perform a REST POST request."""
        return self.session.post(
            self._format_url(url),
            data=json.dumps(payload, ensure_ascii=False).encode('utf-8'),
            headers=self.header,
            verify=self.verify)

    def put(self, url, payload):
        """Perform a REST PUT request."""
        return self.session.put(
            self._format_url(url),
            data=json.dumps(payload, ensure_ascii=False).encode('utf-8'),
            headers=self.header,
            verify=self.verify)

    def delete(self, url):
        """Perform a REST DELETE request."""
        return self.session.delete(
            self._format_url(url),
            headers=self.header,
            verify=self.verify)



"""Utilities for linux filesystems administration"""
import datetime
import itertools
import logging
import re
import shlex
import shutil
import subprocess
import time
from pathlib import Path

LOG = logging.getLogger(__name__)

FSTAB_CONF = '/etc/fstab'
FSTAB_BACKUP_DIR = '/etc/fstab.bkp'
MULTIPATH_CONF = '/etc/multipath.conf'
MULTIPATH_BACKUP_DIR = '/etc/multipath.bkp'
MULTIPATH_CMD = '/sbin/multipath'


class Process():

    def __init__(self):
        pass

    def run(self, command, stdin=subprocess.PIPE, stdout=subprocess.PIPE):
        """
        Runs a Shell command.

        :param command: command to run.
        :param stdin: stdin of the command execution, default: subprocess.PIPE.
        :param stdout: stdout of the command execution, default: subprocess.PIPE.
        :returns: tuple (return_code, stdout_data, stderr_data).
        """
        sub_cmds = [list(x[1]) for x in itertools.groupby(shlex.split(command), lambda x: x == '|') if not x[0]]

        processes = []
        for sub_cmd in sub_cmds:
            p = subprocess.Popen(sub_cmd, stdin=stdin, stdout=stdout)
            stdin = p.stdout
            processes.append(p)

        result = returncode = None
        num_proc = len(processes)
        for idx, proc in enumerate(processes):
            if idx == num_proc - 1:
                result = proc.communicate()
                returncode = proc.returncode
            else:
                proc.stdout.close()

        if returncode != 0:
            LOG.error("Run command: [%s] | Return Code: %s | Stdout: %s | Stderr: %s.",
                      command, returncode, result[0], result[1])
        else:
            LOG.info("Run command: [%s] | Return Code: %s | Number of words in stdout: %s | Stderr: %s.",
                     command, returncode, len(result[0].split()), result[1])

        return (returncode, result[0].decode(encoding='utf-8', errors='strict').rstrip('\n'), result[1])

    def get_hostname(self):
        """
        Gets hostname of local server.

        :returns: server's hostname or `None` for error.
        """
        cmd = 'hostname'
        result = self.run(cmd)

        return result[1] if result[0] == 0 else False

    def mountpoint_exists(self, mountpoint):
        """
        Checks if mountpoint exists.

        :param mountpoint: mountpoint to check.
        :returns: `True` if the mountpoint exists, `False` otherwise.
        """
        if not Path(mountpoint).is_dir():
            LOG.error('%s is not a directory.', mountpoint)
            return False

        cmd = f"df -h | egrep '{mountpoint}' | wc -l"
        result = self.run(cmd)

        return True if int(result[1]) == 1 else False

    def get_fsdevice_from_mountpoint(self, mountpoint):
        """
        Gets filesystem device from mountpoint.

        Example: mountpoint = /d03 --> fsdevice = /dev/mapper/vold03p1.

        :param mountpoint: the mountpoint to look for its filesystem device.
        :returns: fsdevice or `None` if fsdevice not found.
        """
        if not self.mountpoint_exists(mountpoint):
            return None

        # get mountpoint
        cmd = f"df -h | egrep '{mountpoint}' | awk '{{ print $1 }}'"
        result = self.run(cmd)

        return result[1] if result[0] == 0 else None

    def get_mpath_from_mountpoint(self, mountpoint):
        """
        Gets mpath from mountpoint.

        Example: mountpoint = /d03 --> mpath = /dev/mapper/vold03.

        :param mountpoint: the mountpoint to look for its mpath.
        :returns: mpath of mountpoint, or None if no mpath is found.
        """
        partname = self.get_fsdevice_from_mountpoint(mountpoint)

        if partname is not None:
            if partname.endswith('p1'):
                return partname[:-2]
            else:
                LOG.error('mpath of volume mounted %s cannot be found.', mountpoint)
        else:
            LOG.error('Cannot get fs device from the mountpoint %s.', mountpoint)

        return None

    def get_wwn_from_mpath(self, mpath_device):
        """
        Gets `wwn` of a mpath.

        Example: mpath = vold03p1 --> wwn = '6000d31000e3940000000000000001fd'.

        :param mpath_device: mpath device to look for its wwn.
        :returns: wwn of mpath device, `None` otherwise.
        """
        result0, result1 = self.run(f"dirname {mpath_device}"), self.run(f"basename {mpath_device}")

        if 1 in [result0[0], result1[0]]:
            LOG.error('Cannot get dirname (%s) or basename (%s) of mpath %s.', result0[1], result1[1], mpath_device)
            return None

        dirname, basename = result0[1], result1[1]

        if dirname in ['.', '/']:
            dirname = '/dev/mapper'

        mpath = f'{dirname}/{basename}'

        # issue sg_inq command
        cmd = f"/usr/bin/sg_inq -i {mpath}"
        result = self.run(cmd)

        if result[0] != 0:
            return None

        try:
            return re.search(r'\[0x.{32}\]', result[1]).group()[3:35]
        except AttributeError:
            LOG.error('WWN of mpath %s is not found in sg_inq command output.', mpath)

        return None

    def get_mpath_from_wwn(self, wwn):
        """
        Gets `mpath` of a `wwn` by looking in `multipath -ll` output.

        Example: wwn = '6000d31000e3940000000000000001fd' --> mpath = vold03.

        :param mpath: mpath to look for its wwn.
        :returns: mpath or None if mpath is not found.
        """
        if not self.is_wwn_valid(wwn):
            return None

        # show multipath config and get mpath of a wwn
        cmd = f"{MULTIPATH_CMD} -ll | grep -i '{wwn}' | awk '{{ print $1 }}'"
        result = self.run(cmd)

        return result[1] if result[0] == 0 else None

    def is_wwn_valid(self, wwn):
        """
        Checks if `wwn` is valid.

        A valid WWN should have 32 or 33 characters.

        :param wwn: the wwn to check
        :returns: `True` if wwn is valid, otherwise, `False`.
        """
        if len(wwn) not in [32, 33]:
            LOG.error('%s is invalid wwn.', wwn)
            return False
        else:
            return True

    def is_wwn_mounted(self, wwn):
        """
        Checks if a device with `wwn` is mounted.

        :param wwn: `wwn` of the device to check.
        :returns: mountpoint if the device is mounted, `None`, otherwise.
        """
        if not self.is_wwn_valid(wwn):
            return None

        alias = self.get_mpath_from_wwn(wwn)
        if alias is None:
            return None

        # check if wwn is mounted
        cmd = f"df -h | grep '/dev/mapper/{alias}p1' | awk '{{ print $6 }}'"
        result = self.run(cmd)

        if result[0] == 0:
            return None if result[1] == '' else result[1]
        else:
            return None

    def generate_devicemapper_alias(self):
        """
        Generates a new device mapper alias.

        :returns: alias name, `None` for error.
        """
        # get existing aliases in MULTIPATH_CONF
        cmd = f"grep alias {MULTIPATH_CONF} | awk '{{ print $2 }}'"
        result = self.run(cmd)
        if result[0] != 0:
            return None

        lst_aliases = result[1].split()
        alias_pattern = re.compile(r'vold\d\d')
        lst_indexes = [int(alias.replace('vold', '')) for alias in lst_aliases if alias_pattern.match(alias)]
        lst_indexes.sort()

        if len(lst_indexes) == 0:
            return 'vold01'

        new_index = lst_indexes[-1] + 1
        new_alias = f"vold{new_index:02d}"

        # make sure the new alias does not exist in lst_aliases
        assert new_alias not in lst_aliases

        return new_alias

    def is_device_valid(self, device):
        """
        Checks if device name is valid.

        A valid device name should start with '/dev/mapper/vold' and ends with 'p1'.

        :param device: device name to check.
        :returns: `True` if device name is valid, otherwise, `False`.
        """
        device_name_pattern = re.compile(r'^/dev/mapper/vold.*p1$')
        if device_name_pattern.match(device) is None:
            LOG.error('%s is not a valid device name.', device)
            return False

        return True

    def get_wwid_from_wwn(self, wwn):
        """Obtains wwid from wwn by looking at `multipath -ll` command output."""
        wwid_cmd = f"{MULTIPATH_CMD} -ll | grep -i '{wwn}' | awk '{{ print $2 }}' | sed -e 's/(//g' -e 's/)//g'"
        wwid_result = self.run(wwid_cmd)
        if wwid_result[0] != 0 or str(wwid_result[1]) == '':
            LOG.error('WWID cannot be obtained from multipath -ll. Result: %s.', wwid_result)
            return None

        return str(wwid_result[1])

    def add_multipath_alias(self, wwn):
        """
        Adds a multipath alias to MULTIPATH_CONF file.

        :param wwn: wwn to add to MULTIPATH_CONF.
        :returns: `True` for success, `False` for error.
        """
        if not self.is_wwn_valid(wwn):
            return False

        # check if wwn already exists in MULTIPATH_CONF file
        if wwn in self.read_file(MULTIPATH_CONF):
            LOG.warning('WWN %s is already in %s.', wwn, MULTIPATH_CONF)
            return False

        wwid = self.get_wwid_from_wwn(wwn)
        if wwid is None:
            return False

        # generate new alias
        alias = self.generate_devicemapper_alias()
        if alias is None:
            LOG.error('Alias for wwn %s could not be generated.', wwn)
            return False

        # backup MULTIPATH_CONF file
        if not self.backup_file(MULTIPATH_CONF, MULTIPATH_BACKUP_DIR):
            return False

        # create the multipath entry to add to MULTIPATH_CONF file
        new_multipath_entry = f"multipath {{" \
                              f"\n             wwid {wwid}" \
                              f"\n             alias {alias}" \
                              f"\n          }}" \
                              f"\n#MULTIPATH_PLACEHOLDER"

        # replace #MULTIPATH_PLACEHOLDER by new_multipath_entry in MULTIPATH_CONF file
        content = self.read_file(MULTIPATH_CONF)
        config = content.replace('#MULTIPATH_PLACEHOLDER', new_multipath_entry)
        new_config_file = open(MULTIPATH_CONF, 'w')
        new_config_file.write(config)
        new_config_file.close()

        return self.reload_multipathd()

    def remove_multipath_entry(self, wwn, alias):
        """
        Removes a multipath entry from MULTIPATH_CONF file.

        :param wwn: wwn of the multipath entry to remove.
        :param alias: alias of the multipath entry to remove
        :returns: `True` for success, `False` for error.
        """
        # backup MULTIPATH_CONF file
        if not self.backup_file(MULTIPATH_CONF, MULTIPATH_BACKUP_DIR):
            return False

        wwid = self.get_wwid_from_wwn(wwn)
        if wwid is None:
            return False

        content = self.read_file(MULTIPATH_CONF)
        regex = r"multipath\s*{\s*wwid\s+" + re.escape(wwid) + r"\s+alias\s+" + re.escape(alias) + r"\s*}\s*"
        if re.search(regex, content) is None:
            LOG.warning('Multipath entry for wwn %s and alias %s does not exist.', wwid, alias)
            return True

        new_content = re.sub(regex, '', content)
        new_config_file = open(MULTIPATH_CONF, 'w')
        new_config_file.write(new_content)
        new_config_file.close()

        return True

    def mount(self, fs):
        """
        Mounts a filesystem or mountpoint if provided.

        :returns: `True` for success, `False` for error.
        """
        cmd = f'mount {fs}'
        result = self.run(cmd)

        return True if result[0] == 0 else False

    def umount(self, fs):
        """
        Unmounts a filesystem, or mountpoint if provided.

        :returns: `True` for success, `False` for error.
        """
        cmd = f'umount {fs}'
        result = self.run(cmd)

        return True if result[0] == 0 else False

    def is_device_in_fstab(self, device):
        """
        Checks if device is in FSTAB_CONF file.

        :param device: device to check in FSTAB_CONF.
        :returns: `True` if device in FSTAB_CONF, otherwise, `False`.
        """
        if not self.is_device_valid(device):
            return False

        volname = device.split('/')[3]

        # check if a volume name exists in FSTAB_CONF
        cmd = f"cat {FSTAB_CONF} | grep -v '^#' | grep '{volname}' | wc -l"
        result = self.run(cmd)

        return True if (result[0], int(result[1])) == (0, 1) else False

    def is_string_in_file(self, string, filename):
        """
        Checks if a `string` is in `filename`.

        :param string: the string to look for in the file.
        :param file: the file to search in.
        :returns: `True` if `string` in `file`, `False`, otherwise.
        """
        if not Path(filename).exists():
            LOG.error('File %s does not exist.', filename)
            return False

        if not Path(filename).is_file():
            LOG.error('%s is not a file.', filename)
            return False

        if string not in self.read_file(filename):
            LOG.error('%s not in %s.', string, filename)
            return False

        return True

    def replace_entry_in_fstab(self, old_fs_device, new_fs_device):
        """
        Replaces entry in FSTAB_CONF file.

        :param old_fs_device: fs device to be replaced.
        :param new_fs_device: new fs device to replace old_fs_device.
        :return: `True` if task completed successfully, otherwise, `False`.
        """
        if not self.is_device_valid(old_fs_device) or not self.is_device_valid(new_fs_device):
            return False

        if not self.is_string_in_file(old_fs_device, FSTAB_CONF):
            return False

        if not self.backup_file(FSTAB_CONF, FSTAB_BACKUP_DIR):
            return False

        old_vol_name = old_fs_device.split('/')[3]
        new_vol_name = new_fs_device.split('/')[3]

        # replace old volume name with new volume name in FSTAB_CONF
        cmd = f"sed -i 's/{old_vol_name}/{new_vol_name}/' {FSTAB_CONF}"
        result = self.run(cmd)

        return True if result[0] == 0 else False

    def add_entry_in_fstab(self, fs_device, mountpoint, fstype='ext4', opts='_netdev'):
        """
        Adds `fs_device` in FSTAB_CONF file.

        :param fs_device: fs device to add.
        :param mountpoint: mountpoint of fs device.
        :param fstype: fs type of fs device.
        :returns: `True` for success, `False` for error.
        """
        if not self.is_device_valid(fs_device):
            return False

        if fstype not in ['ext3', 'ext4', 'xfs']:
            LOG.error('fstype is not supported.')
            return False

        if not self.backup_file(FSTAB_CONF, FSTAB_BACKUP_DIR):
            return False

        new_fstab_entry = f"{fs_device}\t\t{mountpoint}\t\t{fstype}\t{opts}\t1 1\n"
        with open(FSTAB_CONF, 'a') as fstab_conf_file:
            fstab_conf_file.write(new_fstab_entry)

        return True

    def remove_entry_from_fstab(self, fs_device):
        """
        Removes `fs_device` from FSTAB_CONF file.

        :param fs_device: fs device to remove.
        :returns: `True` for success, `False` for error.
        """
        if not self.is_device_valid(fs_device):
            return False

        if not self.is_string_in_file(fs_device, FSTAB_CONF):
            return False

        if not self.backup_file(FSTAB_CONF, FSTAB_BACKUP_DIR):
            return False

        content = self.read_file(FSTAB_CONF)
        regex = r"\n" + re.escape(fs_device) + r"\s+.*"
        if re.search(regex, content) is None:
            LOG.error('fstab entry for %s does not exist.', fs_device)
            return False

        new_content = re.sub(regex, '', content)
        new_config_file = open(FSTAB_CONF, 'w')
        new_config_file.write(new_content)
        new_config_file.close()

        return True

    def flush_multipath_devices(self):
        """
        Flushes multipath devices via the command `multipath -F`.

        :returns: `True` for success, `False` for error.
        """
        cmd = f'{MULTIPATH_CMD} -F'
        result = self.run(cmd)

        return True if result[0] == 0 else False

    def reload_multipathd(self):
        """
        Reloads the multipathd service.

        :returns: `True` for success, `False` for error.
        """
        cmd = 'service multipathd reload'
        result = self.run(cmd)

        time.sleep(5)

        return True if result[0] == 0 else False

    def daemon_reload(self):
        """
        Reloads systemd manager configuration

        :returns: `True` for success, `False` for error.
        """
        cmd = 'systemctl daemon-reload'
        result = self.run(cmd)

        time.sleep(5)

        return True if result[0] == 0 else False

    def get_multipath_raw_devices(self, mpath_alias):
        """
        Gets multipath raw devices of a multipath device.

        Example: mpath_alias = vold01 --> ['sdc', 'sdb'].

        :param mpath_alias: multipath device alias.
        :returns: list of multipath raw devices, `None` for error.
        """
        cmd = f"{MULTIPATH_CMD} -ll {mpath_alias} | cut -c6- | grep -E 'ready|faulty' | awk '{{ print $2 }}'"
        result = self.run(cmd)

        return result[1].split('\n') if result[0] == 0 and result[1] != '' else None

    def delete_raw_devices(self, raw_device):
        """
        Deletes a raw device.

        :param raw_device: raw device to delete.
        :returns: `True` for success, `False` for error.
        """
        with open(f'/sys/block/{raw_device}/device/delete', 'w') as outfile:
            outfile.write('1')

        return True

    def is_dir_empty(self, path):
        """
        Checks if a directory is empty.

        :param path: path of the directory to check.
        :returns: `True` if the directory, `False` otherwise.
        """
        if not Path(path).is_dir():
            LOG.error('Path %s is not a directory.', path)
            return False

        if len(list(Path(path).glob('*'))) > 0:
            LOG.error('Directory %s is not empty.', path)
            return False

        return True

    def rescan_scsibus(self):
        """
        Scans SCSI bus on local server.

        :return: `True` for success, `False` on error.
        """
        for host in Path('/sys/class/scsi_host').iterdir():
            with open(f'{host}/scan', 'w') as outfile:
                outfile.write('- - -\n')

        # wait for 20 seconds
        time.sleep(20)

        LOG.info('rescan_scsibus() succeeded.')

        # TODO we need to check if file writing succeeded
        return True

    def backup_file(self, file_, backup_dir):
        """
        Backups a file.

        Copies `file_` to `backup_dir` directory and name the new file `file.timestamp`.

        :param file_: the full path and the name of the file to back up.
        :param backup_dir: the path of the new file.
        :returns: `True` if backup task succeeded, `False`, otherwise.
        """
        timestamp = datetime.datetime.now().strftime('%Y%m%d_%H:%M:%S')

        if not Path(file_).exists():
            LOG.error('%s does not exist.', file_)
            return False

        if not Path(file_).is_file():
            LOG.error('%s is not a file.', file_)
            return False

        Path(backup_dir).mkdir(parents=True, exist_ok=True)

        file_basename = Path(file_).resolve().name
        destination = Path(f'{backup_dir}/{file_basename}.{timestamp}')

        try:
            result = shutil.copy(file_, destination)
        except OSError as error:
            LOG.error('File %s cannot be backed up. Error: %s', file_, error)
            return False

        if result == destination:
            LOG.info('backup_file(%s, %s) succeeded.', file_, backup_dir)
            return True
        else:
            LOG.error('backup_file(%s, %s) failed.', file_, backup_dir)
            return False

    def is_snapshot_label_valid(self, label):
        """
        Checks if snapshot label is valid.

        A snapshot label should be a string of alphanumeric characters with a length in [2, 30].

        :param label: the snapshot label to check.
        :return: `True` if the label is valid, `False` otherwise.
        """
        label_pattern = re.compile(r'^\w{2,30}$')
        if label_pattern.match(label) is None:
            LOG.error('%s is not a valid snapshot label.', label)
            return False

        return True

    def record_config_details(self):
        """
        Logs `df`, `multipath`, and `fstab` config details.
        """
        fstab_conf_content = self.read_file(FSTAB_CONF)
        multipath_conf_content = self.read_file(MULTIPATH_CONF)

        df_cmd_result = self.run('df -h')
        df_output = df_cmd_result[1]

        multipath_cmd_result = self.run('multipath -ll')
        multipath_output = multipath_cmd_result[1]

        LOG.info('Content of /etc/fstab:\n%s', fstab_conf_content)
        LOG.info('Content of /etc/multipath.conf:\n%s', multipath_conf_content)

        LOG.info('Output of df -h:\n%s', df_output)
        LOG.info('Output of multipath -ll:\n%s', multipath_output)

    def read_file(self, filename):
        """
        Reads a file.

        :param filename: the name of file to read.
        :returns: content of the file to read, or `None` on error.
        """
        content = None
        try:
            content = open(filename).read()
        except Exception as e:
            LOG.error('Cannot read %s. Error: %s', filename, e)

        return content



import logging
import sys
from logging.handlers import RotatingFileHandler
from pathlib import Path

from ucamdsm.lib.options import Options

LOG = logging.getLogger(__name__)

LOG_DIR = '/var/log/ucamdsm'
LOG_FILE = 'ucamdsm.log'
ROTATING_FILE_HANDLER_MAXBYTES = 10 * 1000 * 1000  # 10 mbytes
ROTATING_FILE_HANDLER_BACKUPCOUNT = 5

Path(LOG_DIR).mkdir(parents=True, exist_ok=True)

stdout_handler = logging.StreamHandler(sys.stdout)
file_handler = RotatingFileHandler(filename=f'{LOG_DIR}/{LOG_FILE}',
                                   maxBytes=ROTATING_FILE_HANDLER_MAXBYTES,
                                   backupCount=ROTATING_FILE_HANDLER_BACKUPCOUNT)

logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s [%(levelname)s] %(message)s',
                    datefmt='%m/%d/%Y %H:%M:%S',
                    handlers=[file_handler, stdout_handler])


def run_project(args):
    options = Options()
    return options.parse(args[1:])


def main():
    result = False
    try:
        result = run_project(sys.argv)
    except KeyboardInterrupt:
        sys.exit()
    except Exception as error:
        LOG.error('Please check your command input. %s', error)
        sys.exit()

    if result:
        LOG.info('[Command succeeded - Returns %s]', result)
    else:
        LOG.error('[Command failed - Returns %s]', result)

    return result


if __name__ == '__main__':
    main()



import setuptools

setuptools.setup(
     name='ucamdsm',
     version='1.2.15',
     packages=['ucamdsm', 'ucamdsm.lib', ],
     entry_points={
        'console_scripts': [
            'ucamdsm = ucamdsm.__main__:main'
        ]
     },
     scripts=['dellsc/replace_volume', 'dellsc/delete_volume',
              'dellsc/create_snapshot', 'dellsc/clone_volume',
              'dellsc/record_recycled_volumes', 'dellsc/remove_recycled_volumes', ],
     license="MIT License (MIT)",
     author="UIS Infrastructure Servers and Storage, University of Cambridge",
     author_email="wh330@cam.ac.uk",
     description="A Dell SC tasks automation using REST API",
     long_description=open('README.md').read(),
     long_description_content_type='text/markdown',
     url="https://gitlab.developers.cam.ac.uk/uis/infra/dell-sc",
)



"""
Functions to manipulate objects on Dell Storage Manager

REST API documentation:
https://downloads.dell.com/Manuals/Common/dell-storage-manager-rest-api-cookbook-3029-wp-san_en-us.pdf
"""
import logging

from .httpclient import HttpClient

LOG = logging.getLogger(__name__)


class DSMConnection():
    """
    Storage Center API interface.

    Handles calls to Dell Storage Manager via the REST API interface.
    """
    def __init__(self, host, port, user, password, verify):
        """
        This creates a connection to Dell Storage Manager.

        :param host: IP address of the Dell Data Collector.
        :param port: Port the Data Collector is listening on.
        :param user: User account to login with.
        :param password: Password.
        :param verify: Boolean indicating whether certificate verification should be turned on or not.
        """
        self.notes = 'Created via Dell Storage Manager REST API'
        self.client = HttpClient(host, port, user, password, verify)
        self.connection_id = None

        self.scs = {}
        self.servers = {}
        self.volumes = {}
        self.volume_folders = {}
        self.replays = {}
        self.rply_profiles = {}
        self.recycled_vols = {}

    def __enter__(self):
        self.open_connection()
        return self

    def __exit__(self, tipe, value, traceback):
        self.close_connection()
        self.client = None

    def _check_result(self, rest_response):
        """
        Checks and logs API responses.

        :param rest_response: The result from a REST API call.
        :returns: ``True`` if success, ``False`` otherwise.
        """
        if 200 <= rest_response.status_code < 300:
            # API call was a normal success
            return True

        LOG.error('REST call result:\tCode: %s\tReason: %s\tText: %s',
                  rest_response.status_code, rest_response.reason, rest_response.text)

        return False

    def _get_json(self, blob):
        """
        Returns a dict from the JSON of a REST response.

        :param blob: The response from a REST call.
        :returns: JSON or None on error.
        """
        try:
            return blob.json()
        except AttributeError:
            LOG.error('Error invalid json: %s', blob)
        return None

    def _get_id(self, blob):
        """
        Returns the instanceId from a Dell REST object.

        :param blob: A Dell SC REST call's response.
        :returns: The instanceId from the Dell SC object or None on error.
        """
        try:
            if isinstance(blob, dict):
                return blob.get('instanceId')
        except AttributeError:
            LOG.error('Invalid API object: %s', blob)
        return None

    def open_connection(self):
        """Authenticate against Dell Storage Manager."""
        payload = {}
        response = self.client.post('ApiConnection/Login', payload)
        if not self._check_result(response):
            LOG.error('Failed to connect to Storage Manager. Response: %s.', response.text)
            return False

        try:
            # log in and set the connection ID
            apidict = self._get_json(response)
            instanceId, version = apidict['instanceId'], apidict['apiVersion']
            LOG.info('Connected to DSM. API Version: %s, Connection ID: %s' % (version, instanceId))
            self.connection_id = self._get_id(apidict)
        except Exception:
            LOG.error('Unrecognized Login Response: %s.', response.text)

    def close_connection(self):
        """Logout of Dell Storage Manager."""
        response = self.client.post('ApiConnection/Logout', {})
        result = self._check_result(response)
        LOG.info('Logout of DSM: %s.', result)

    def get_scs(self):
        """
        Gets all SCs and updates `self.scs`.

        :returns: `True` for success, `False` on error.
        """
        path = 'ApiConnection/ApiConnection/%s/StorageCenterList' % self.connection_id
        response = self.client.get(path)
        if not self._check_result(response):
            LOG.error('Failed to list Storage Centers. Response: %s', response.text)
            return False

        self.scs = {}
        stdout = self._get_json(response)
        for i in range(len(stdout)):
            self.scs[stdout[i]['name']] = {}
            self.scs[stdout[i]['name']]['instanceId'] = stdout[i]['instanceId']
            self.scs[stdout[i]['name']]['hostOrIP'] = stdout[i]['hostOrIpAddress']

        LOG.info('API call %s succeeded.', path)

        return True

    def get_volume_folders(self):
        """
        Gets volume folders in all SCs and updates `self.volume_folders`.

        :returns: `True` for success, `False` on error.
        """
        results = []
        self.volume_folders = {}

        results.append(self.get_scs())
        for sc in self.scs.values():
            sc_id = sc['instanceId']
            result = self.get_volume_folders_in_sc(sc_id)
            results.append(result)

        LOG.info('get_volume_folders() result: %s.', results)

        return False if False in results else True

    def get_volume_folders_in_sc(self, sc_id):
        """
        Gets volume folders in a given SC and updates `self.volume_folders`.

        :param sc_id: Storage Center instance ID.
        :returns: dict object containing the retrieved volume folders.
        """
        path = 'StorageCenter/StorageCenter/%s/VolumeFolderList' % sc_id
        response = self.client.get(path)
        if not self._check_result(response):
            LOG.error('Failed to get volume folders in SC %s. Response: %s', sc_id, response.text)
            return False

        stdout = self._get_json(response)
        for i in range(len(stdout)):
            self.volume_folders[stdout[i]['instanceId']] = {}
            self.volume_folders[stdout[i]['instanceId']]['name'] = stdout[i]['name']
            if stdout[i].get('parent'):
                self.volume_folders[stdout[i]['instanceId']]['parent'] = stdout[i]['parent']['instanceName']
            self.volume_folders[stdout[i]['instanceId']]['scName'] = stdout[i]['scName']

        LOG.info('API call %s succeeded.', path)

        return True

    def get_volumes(self):
        """
        Gets volumes in all SCs and updates `self.volumes`.

        :returns: `True` for success, `False` on error.
        """
        results = []
        self.volumes = {}

        results.append(self.get_scs())
        for sc in self.scs.values():
            sc_id = sc['instanceId']
            result = self.get_volumes_in_sc(sc_id)
            results.append(result)

        LOG.info('get_volumes() results %s.', results)

        return False if False in results else True

    def get_volumes_in_sc(self, sc_id):
        """
        Gets volumes in a given SC and updates `self.volumes`.

        :param sc_id: Storage Center instance ID.
        :returns: dict object containing the retrieved volumes.
        """
        path = 'StorageCenter/StorageCenter/%s/VolumeList' % sc_id
        response = self.client.get(path)
        if not self._check_result(response):
            LOG.error('Failed to get volumes in SC %s. Response: %s', sc_id, response.text)
            return False

        stdout = self._get_json(response)
        for i in range(len(stdout)):
            self.volumes[stdout[i]['deviceId']] = {}
            self.volumes[stdout[i]['deviceId']]['instanceId'] = stdout[i]['instanceId']
            self.volumes[stdout[i]['deviceId']]['inRecycleBin'] = stdout[i]['inRecycleBin']
            self.volumes[stdout[i]['deviceId']]['volumeFolderPath'] = stdout[i]['volumeFolderPath']
            self.volumes[stdout[i]['deviceId']]['volumeFolderName'] = stdout[i]['volumeFolder']['instanceName']
            self.volumes[stdout[i]['deviceId']]['volumeFolderId'] = stdout[i]['volumeFolder']['instanceId']
            self.volumes[stdout[i]['deviceId']]['status'] = stdout[i]['status']
            self.volumes[stdout[i]['deviceId']]['name'] = stdout[i]['name']
            self.volumes[stdout[i]['deviceId']]['scName'] = stdout[i]['scName']

        LOG.info('API call %s succeeded.', path)

        return True

    def get_servers(self):
        """
        Gets servers in all SCs and updates `self.servers`.

        :returns: `True` for success, `False` on error.
        """
        results = []
        self.servers = {}

        results.append(self.get_scs())
        for sc in self.scs.values():
            sc_id = sc['instanceId']
            result = self.get_servers_in_sc(sc_id)
            results.append(result)

        LOG.info('get_servers() result: %s.', results)

        return False if False in results else True

    def get_servers_in_sc(self, sc_id):
        """
        Gets servers in a given SC and updates `self.servers`.

        :param sc_id: Storage Center instance ID.
        :returns: `True` for success, `False` for error.
        """
        path = 'StorageCenter/StorageCenter/%s/ServerList' % sc_id
        response = self.client.get(path)
        if not self._check_result(response):
            LOG.error('Failed to get servers in SC %s. Response: %s.', sc_id, response.text)
            return False

        stdout = self._get_json(response)
        for i in range(len(stdout)):
            self.servers[stdout[i]['name']] = {}
            self.servers[stdout[i]['name']]['instanceId'] = stdout[i]['instanceId']
            self.servers[stdout[i]['name']]['serverFolderPath'] = stdout[i]['serverFolderPath']
            self.servers[stdout[i]['name']]['status'] = stdout[i]['status']
            self.servers[stdout[i]['name']]['operatingSystem'] = stdout[i]['operatingSystem']['instanceName']
            self.servers[stdout[i]['name']]['scName'] = stdout[i]['scName']

        LOG.info('API call %s succeeded.', path)

        return True

    def get_snapshots(self):
        """
        Gets snapshots in all SCs and updates `self.replays`.

        :param sc_id: Storage Center instance ID.
        :returns: `True` for success, `False` for error.
        """
        payload = {}
        path = 'StorageCenter/ScReplay/GetList'
        response = self.client.post(path, payload)
        if not self._check_result(response):
            LOG.error('Failed to list snapshots. Response: %s', response.text)
            return False

        self.replays = {}
        stdout = self._get_json(response)
        for i in range(len(stdout)):
            # excluding snapshots with no createVolume attributes
            if stdout[i].get('createVolume') is not None:
                self.replays[stdout[i]['createVolume']['instanceName']] = {}
                self.replays[stdout[i]['createVolume']['instanceName']]['volumeId'] = \
                    stdout[i]['createVolume']['instanceId']
                self.replays[stdout[i]['createVolume']['instanceName']]['instanceId'] = stdout[i]['instanceId']
                self.replays[stdout[i]['createVolume']['instanceName']]['description'] = stdout[i]['description']
                self.replays[stdout[i]['createVolume']['instanceName']]['scName'] = stdout[i]['scName']

        LOG.info('API call %s succeeded.', path)

        return True

    def get_snapshot_profiles(self):
        """
        Gets snapshot profiles, aka replay profiles, in all SCs and updates `self.rply_profiles`.

        :returns: `True` for success, `False` for error.
        """
        payload = {}
        path = 'StorageCenter/ScReplayProfile/GetList'
        response = self.client.post(path, payload)
        if not self._check_result(response):
            LOG.error('Failed to list reply profiles. Response: %s.', response.text)
            return False

        self.rply_profiles = {}
        stdout = self._get_json(response)
        for i in range(len(stdout)):
            self.rply_profiles[stdout[i]['instanceName']] = {}
            self.rply_profiles[stdout[i]['instanceName']]['instanceId'] = stdout[i]['instanceId']
            self.rply_profiles[stdout[i]['instanceName']]['scName'] = stdout[i]['scName']

        LOG.info('API call %s succeeded.', path)

        return True

    def create_snapshot(self, volume_id, description=None, expiration='5'):
        """
        Creates a snapshot from volume.

        :param volume: instance ID of the volume from which to create a snapshot.
        :param name: name of the snapshot to create.
        :param description: description of the snapshot.
        :param expiration: expiration time in minutes, default: 5 minutes.
        :returns: created snapshot ID, `None` on error.
        """
        payload = {}
        if description is None:
            description = 'Snapshot created via REST API'
        payload['Description'] = description
        payload['ExpireTime'] = str(expiration)

        path = '/StorageCenter/ScVolume/%s/CreateReplay' % volume_id
        response = self.client.post(path, payload)
        if not self._check_result(response):
            LOG.error('Failed to create snapshot for volume ID %s. Response: %s.', volume_id, response.text)
            return None

        stdout = self._get_json(response)

        LOG.info('API call %s succeeded.', path)

        return self._get_id(stdout)

    def create_view_volume(self, snapshot_id, replay_profile_id, volume_name, volume_folder_id):
        """
        Creates a view volume from a snapshot.

        :param snapshot_id: ID of snapshot from which to create the view volume.
        :param replay_profile_id: ID of the replay profile for the view volume.
        :param volume_name: name to assign to newly created volume.
        :param volume_folder_id: instance ID of the folder of the new volume.
        :returns: view volume ID, `None` on error.
        """
        payload = {}
        payload['Name'] = volume_name
        payload['VolumeFolder'] = volume_folder_id
        payload['Notes'] = 'View volume created via REST API'

        if replay_profile_id is not None:
            payload['ReplayProfileList'] = [replay_profile_id]

        path = 'StorageCenter/ScReplay/%s/CreateView' % snapshot_id
        response = self.client.post(path, payload)
        if not self._check_result(response):
            LOG.error('Failed to create view volume %s. Response: %s.', volume_name, response.text)
            return None

        stdout = self._get_json(response)

        LOG.info('API call %s succeeded.', path)

        return self._get_id(stdout)

    def map_volume_to_server(self, volume_id, server_id):
        """
        Maps a volume to a server.

        :param volume_id: ID of the volume to map.
        :param server_id: ID of the server to which to map the volume.
        :returns: `True` for success, `False` for error.
        """
        payload = {}
        payload['Server'] = server_id

        path = 'StorageCenter/ScVolume/%s/MapToServer' % volume_id
        response = self.client.post(path, payload)
        if not self._check_result(response):
            LOG.error('Failed to map volume ID %s to server ID %s. Response: %s.',
                      volume_id, server_id, response.text)
            return False

        LOG.info('API call %s succeeded.', path)

        return True

    def unmap_volume(self, volume_id):
        """
        Unmap a volume.

        :param volume_id: ID of the volume to unmap.
        :returns: `True` for success, `False` for error.
        """
        payload = {}

        path = 'StorageCenter/ScVolume/%s/Unmap' % volume_id
        response = self.client.post(path, payload)
        if not self._check_result(response):
            LOG.error('Failed to unmap volume with ID %s. Response: %s.', volume_id, response.text)
            return False

        LOG.info('API call %s succeeded.', path)

        return True

    def delete_volume(self, volume_id, recycle=True):
        """
        Deletes a volume.

        :param volume_id: ID of the volume to delete.
        :param recycle: If `True`, move the volume to recycle bin, otherwise, delete the volume PERMANENTLY.
        :returns: `True` for success, `False` for error.
        """
        payload = {}

        if recycle:
            path = 'StorageCenter/ScVolume/%s/Recycle' % volume_id
            response = self.client.post(path, payload)
        else:
            path = 'StorageCenter/ScVolume/%s' % volume_id
            response = self.client.delete(path)

        if not self._check_result(response):
            LOG.error('Failed to delete volume with ID %s. Response: %s.', volume_id, response.text)
            return False

        LOG.info('API call %s succeeded.', path)

        return True

    def get_vol_mappings(self, volume_id):
        """
        Gets volume mappings.

        :param volume_id: ID of volume to look for its mappings.
        :returns: mappings dict, or None on error.
        """
        path = 'StorageCenter/ScVolume/%s/MappingList' % volume_id
        response = self.client.get(path)
        if not self._check_result(response):
            LOG.error('Failed to get volume mappings of volume with ID %s. Response: %s.', volume_id, response.text)
            return None

        stdout = self._get_json(response)

        LOG.info('API call %s succeeded.', path)

        return stdout

    def get_recycled_volumes(self):
        """
        Gets recycled volumes and updates `self.recycled_vols`.

        :retrun: `True` for success, `False` on error.
        """
        result = self.get_volumes()
        self.recycled_vols = {}

        for wwn, vol_info in self.volumes.items():
            # we're sure 'inRecycleBin' is in self.volumes[wwn]
            if self.volumes[wwn]['inRecycleBin']:
                self.recycled_vols[wwn] = vol_info

        LOG.info('get_recycled_volumes() result %s.', result)

        return result

    def get_server_id(self, server_name):
        """
        Returns server ID from server name.

        :param server_name: the name of server to return its ID.
        :returns: server ID or `None` on failure.
        """
        self.get_servers()
        if self.servers.get(server_name) is None:
            LOG.error('Server %s not found on Dell SCs.', server_name)
            return None

        # TODO: is server name unique to use it to determine the server id confidently?
        server_id = self.servers[server_name]['instanceId']

        return server_id



import argparse
import getpass
import json
import logging
from argparse import ArgumentParser, ArgumentTypeError

from .connection import DSMConnection
from .process import Process
from .task import Task

LOG = logging.getLogger(__name__)


class Options:

    def __init__(self):
        self.process = Process()
        self._init_parser()

    def _init_parser(self):
        """Define arguments for the CLI."""
        self.parser = ArgumentParser(description='Manage Dell Storage Manager objects via REST API.')

        self.parser.add_argument('--dsm_host',
                                 default='sc-data-ma.admin.cam.ac.uk', type=str,
                                 help='DSM hostname. Default: sc-data-ma.admin.cam.ac.uk')
        self.parser.add_argument('--dsm_port',
                                 default=3033,
                                 help='DSM port. Default: 3033')
        self.parser.add_argument('--dsm_user',
                                 type=str,
                                 help='DSM Data Collector username')
        self.parser.add_argument('--dsm_password',
                                 type=str,
                                 help='DSM Data Collector password')
        self.parser.add_argument('--is_secure',
                                 type=str2bool,
                                 default=False,
                                 help='Secure connection. Default: False')
        self.parser.add_argument('--record_config',
                                 type=str2bool,
                                 default=False,
                                 help='Record multipath and file system config details.\
                                 Default: False (do not record system config details)')
        self.parser.add_argument('--file',
                                 type=open,
                                 action=LoadFromFile,
                                 help='Read arguments from json file')

        self.subparsers = self.parser.add_subparsers(dest='subparser')

        # list dell sc objects
        self.parser_list = self.subparsers.add_parser('list',
                                                      help='List Dell SC objects. \
                                                      Example: ucamdsm --dsm_user user --dsm_password abc123 list \
                                                      --object scs')
        self.parser_list.add_argument(
            '--object',
            required=True,
            choices=['scs', 'servers', 'volumes', 'snapshots', 'volFolders', 'snapshotProfiles', 'recycledVols'])
        self.parser_list.add_argument('--output',
                                      choices=['json'],
                                      help='Format of the output. Supported formats: json')

        # delete_volume
        self.parser_delete_volume = self.subparsers.add_parser('delete_volume',
                                                               help='Delete Dell SC Volume. \
                                                               Example: ucamdsm --dsm_user user \
                                                               --dsm_password abc123 delete_volume \
                                                               --vol_wwn 12345')
        self.parser_delete_volume.add_argument('--vol_wwn',
                                               required=True,
                                               help='ID of the volume to delete. \
                                               Must be mapped to local server')
        self.parser_delete_volume.add_argument('--recycle',
                                               type=str2bool,
                                               default=True,
                                               help='Recycle the volume or delete it permanently. \
                                               Default: True (recycle the volume)')

        # replace_volume
        self.parser_replace_volume = self.subparsers.add_parser('replace_volume',
                                                                help='Clone the volume mounted on source \
                                                                mountpoint and mount it on destination mountpoint. \
                                                                Then, delete the volume was initially mounted on \
                                                                destination mountpoint. Example: ucamdsm --dsm_user \
                                                                user  --dsm_password abc123 replace_volume \
                                                                --src_mp /src --dst_mp /dst')
        self.parser_replace_volume.add_argument('--src_mp',
                                                required=True,
                                                help='Source mountpoint. Must be on local server')
        self.parser_replace_volume.add_argument('--dst_mp',
                                                required=True,
                                                help='Destination mountpoint. Must be on local server')

        # clone_volume
        self.parser_clone_volume = self.subparsers.add_parser('clone_volume',
                                                              help='Create a snapshot with label of a volume \
                                                              with WWN, create a view volume from the snapshot, \
                                                              then, map the view volume to local server and \
                                                              mount it on target mountpoint. Example: ucamdsm \
                                                              --dsm_user user --dsm_password abc123 clone_volume \
                                                              --vol_wwn 12345 --replay_label test1 --target_mp /target')
        self.parser_clone_volume.add_argument('--vol_wwn',
                                              required=True,
                                              help='WWN of the volume from which to create the snapshot')
        self.parser_clone_volume.add_argument('--replay_label',
                                              required=True,
                                              help='Snapshot label.')
        self.parser_clone_volume.add_argument('--target_mp',
                                              required=True,
                                              help='The mountpoint that the view volume will be mounted on. \
                                              Must be on local server')

        # create_snapshot
        self.parser_create_snapshot = self.subparsers.add_parser('create_snapshot',
                                                                 help='Create a snapshot of the volume mounted on \
                                                                 a given mountpoint. Example: ucamdsm --dsm_user \
                                                                 user --dsm_password abc123 create_snapshot --mp /mp \
                                                                 --replay_label test1 --retention 5')
        self.parser_create_snapshot.add_argument('--mp',
                                                 required=True,
                                                 help='The mountpoint where the volume is mounted. \
                                                 Must be on local server')
        self.parser_create_snapshot.add_argument('--replay_label',
                                                 required=True,
                                                 help='Snapshot label')
        self.parser_create_snapshot.add_argument('--retention',
                                                 required=True,
                                                 help='Retention period of the snapshot to create, in days')

        # deleted the recycled vols whose wwns are listed in a file
        self.parser_delete_recycled_vols = self.subparsers.add_parser('delete_recycled_vols',
                                                                      help='Delete recycled volumes whose WWNs are listed in a file. \
                                                                      Example: ucamdsm --dsm_user user --dsm_password \
                                                                      abc123 delete_recycled_vols --wwns_file \
                                                                      /tmp/recycled_vols_wwns.txt')
        self.parser_delete_recycled_vols.add_argument('--wwns_file',
                                                      required=True,
                                                      help='File containing the WWNs of the recycled volumes \
                                                      to be deleted.')

        # map volume to local server
        self.parser_map_volume = self.subparsers.add_parser('map_volume',
                                                            help='Maps the volume to local server and \
                                                            mounts it on the specified mount point. \
                                                            Example: ucamdsm --dsm_user \
                                                            user --dsm_password abc123 map_volume --wwn vol_wwn \
                                                            --mp /d10')
        self.parser_map_volume.add_argument('--wwn',
                                            required=True,
                                            help='WWN of the volume to map to local server.')
        self.parser_map_volume.add_argument('--mp',
                                            required=True,
                                            help='Path of the mountpoint to mount the volume on.')

        # unmap volume from local server
        self.parser_unmap_volume = self.subparsers.add_parser('unmap_volume',
                                                              help='Unmaps the volume from local server. Volume \
                                                              must be unmounted. Example: ucamdsm --dsm_user \
                                                              user --dsm_password abc123 create_snapshot --wwn vol_wwn')
        self.parser_unmap_volume.add_argument('--wwn',
                                              required=True,
                                              help='WWN of the volume to unmap from local server.')

    def parse(self, args=None):
        """Parse given arguments and execute an action accordingly."""
        result = None
        self.known, self.unknown = self.parser.parse_known_args(args)[:]

        if len(self.unknown) != 0:
            LOG.error('*WARN* Unknown args received: %s.', repr(self.unknown))
            return False

        if self.known.subparser is None or \
                self.known.subparser not in ['list', 'delete_volume',
                                             'replace_volume', 'clone_volume', 'create_snapshot',
                                             'delete_recycled_vols']:
            LOG.error('No subcommand is specified (%s).', self.known.subparser)
            return False

        # Read crendential
        if self.known.dsm_user is None:
            self.known.dsm_user = input('Username: ')
        if self.known.dsm_password is None:
            self.known.dsm_password = getpass.getpass()

        # Record config if enabled
        if self.known.record_config:
            self.process.record_config_details()

        if self.known.subparser == 'list':
            result = list_object(self.known)
        elif self.known.subparser == 'delete_volume':
            result = delete_volume(self.known)
        elif self.known.subparser == 'replace_volume':
            result = replace_volume(self.known)
        elif self.known.subparser == 'clone_volume':
            result = clone_volume(self.known)
        elif self.known.subparser == 'create_snapshot':
            result = create_snapshot(self.known)
        elif self.known.subparser == 'delete_recycled_vols':
            result = delete_recycled_vols(self.known)
        elif self.known.subparser == 'map_volume':
            result = map_volume(self.known)
        elif self.known.subparser == 'unmap_volume':
            result = unmap_volume(self.known)

        # Record config if enabled
        if self.known.record_config:
            self.process.record_config_details()

        return result


class LoadFromFile(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        with values as f:
            args = json.load(f)

        for k, v in args.items():
            setattr(namespace, k, v)


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise ArgumentTypeError('Boolean value expected.')


def list_object(args):
    """
    List Dell SC objects.

    Example: ucamdsm --dsm_user user --dsm_password abc123 list --object scs
    """
    with DSMConnection(args.dsm_host, args.dsm_port, args.dsm_user,
                       args.dsm_password, args.is_secure) as dsm_conn:
        result = None
        dsm_objects = None
        if args.object == 'scs':
            result = dsm_conn.get_scs()
            dsm_objects = dsm_conn.scs
        elif args.object == 'servers':
            result = dsm_conn.get_servers()
            dsm_objects = dsm_conn.servers
        elif args.object == 'volumes':
            result = dsm_conn.get_volumes()
            dsm_objects = dsm_conn.volumes
        elif args.object == 'snapshots':
            result = dsm_conn.get_snapshots()
            dsm_objects = dsm_conn.replays
        elif args.object == 'volFolders':
            result = dsm_conn.get_volume_folders()
            dsm_objects = dsm_conn.volume_folders
        elif args.object == 'snapshotProfiles':
            result = dsm_conn.get_snapshot_profiles()
            dsm_objects = dsm_conn.rply_profiles
        elif args.object == 'recycledVols':
            result = dsm_conn.get_recycled_volumes()
            dsm_objects = dsm_conn.recycled_vols

        if result and dsm_objects is not None:
            if args.output is not None and args.output == 'json':
                LOG.info('List of %s:\n%s',
                         args.object, json.dumps(dsm_objects, indent=4, sort_keys=False))
            else:
                LOG.info('List of %s:\n%s',
                         args.object, '\n'.join(str(item) for item in dsm_objects.items()))

    return result


def delete_volume(args):
    """
    Example: ucamdsm --dsm_user user --dsm_password abc123 delete_volume --vol_wwn 12345
    """
    with Task(args.dsm_host, args.dsm_port, args.dsm_user, args.dsm_password, args.is_secure) as task:
        result = task.delete_volume(args.vol_wwn, recycle=args.recycle)

    if result:
        LOG.info('delete_volume(%s, recycle=%s) succeeded. Result: %s.',
                 args.vol_wwn, args.recycle, result)
    else:
        LOG.error('delete_volume(%s, recycle=%s) failed. Result: %s.',
                  args.vol_wwn, args.recycle, result)

    return result


def replace_volume(args):
    """
    Example: ucamdsm --dsm_user user --dsm_password abc123 replace_volume --src_mp /src --dst_mp /dst
    """
    with Task(args.dsm_host, args.dsm_port, args.dsm_user, args.dsm_password, args.is_secure) as task:
        result = task.replace_volume(args.src_mp, args.dst_mp)

    if result:
        LOG.info('replace_volume(%s, %s) succeeded. Result: %s.', args.src_mp, args.dst_mp, result)
    else:
        LOG.error('replace_volume(%s, %s) failed. Result: %s.', args.src_mp, args.dst_mp, result)

    return result


def clone_volume(args):
    """
    Example: ucamdsm --dsm_user user --dsm_password abc123 clone_volume \
        --vol_wwn 12345 --replay_label test1 \
        --target_mp /target
    """
    with Task(args.dsm_host, args.dsm_port, args.dsm_user, args.dsm_password, args.is_secure) as task:
        result = task.clone_volume(args.vol_wwn, args.replay_label, args.target_mp)

    if result:
        LOG.info('clone_volume(%s, %s, %s) succeeded. Result: %s.',
                 args.vol_wwn, args.replay_label, args.target_mp, result)
    else:
        LOG.error('clone_volume(%s, %s, %s) failed. Result: %s.',
                  args.vol_wwn, args.replay_label, args.target_mp, result)

    return result


def create_snapshot(args):
    """
    Example: ucamdsm --dsm_user user --dsm_password abc123 \
         create_snapshot --mp /mp --replay_label test1 --retention 5
    """
    with Task(args.dsm_host, args.dsm_port, args.dsm_user, args.dsm_password, args.is_secure) as task:
        result = task.create_snapshot(args.mp, args.replay_label, args.retention)

    if result:
        LOG.info('create_snapshot(%s, %s, %s) succeeded. Result: %s.',
                 args.mp, args.replay_label, args.retention, result)
    else:
        LOG.error('create_snapshot(%s, %s, %s) failed. Result: %s.',
                  args.mp, args.replay_label, args.retention, result)

    return result


def delete_recycled_vols(args):
    """
    Example: ucamdsm --dsm_user user --dsm_password abc123 \
         delete_recycled_vols --wwns_file /tmp/recycled_vols_wwns.txt
    """
    with Task(args.dsm_host, args.dsm_port, args.dsm_user, args.dsm_password, args.is_secure) as task:
        result = task.delete_recycled_vols(args.wwns_file)

    if result:
        LOG.info('delete_recycled_vols(%s) succeeded. Result: %s.', args.wwns_file, result)
    else:
        LOG.error('delete_recycled_vols(%s failed. Result: %s.', args.wwns_file, result)

    return result


def map_volume(args):
    """
    Example: ucamdsm --dsm_user user --dsm_password abc123 \
         map_volume --wwn 012345 --mp /d10
    """
    with Task(args.dsm_host, args.dsm_port, args.dsm_user, args.dsm_password, args.is_secure) as task:
        result = task.map_dellsc_volume(args.wwn, args.mp)

    if result:
        LOG.info('map_volume(%s, %s) succeeded. Result: %s.', args.wwn, args.mp, result)
    else:
        LOG.error('map_volume(%s, %s) failed. Result: %s.', args.wwn, args.mp, result)

    return result


def unmap_volume(args):
    """
    Example: ucamdsm --dsm_user user --dsm_password abc123 \
         unmap_volume --wwn 012345
    """
    with Task(args.dsm_host, args.dsm_port, args.dsm_user, args.dsm_password, args.is_secure) as task:
        result = task.unmap_dellsc_volume(args.wwn)

    if result:
        LOG.info('delete_recycled_vols(%s) succeeded. Result: %s.', args.wwn, result)
    else:
        LOG.error('delete_recycled_vols(%s failed. Result: %s.', args.wwn, result)

    return result
